from . import settings
from scarlet.utils.lookups import get_model_from_setting


def get_page_model():
    return get_model_from_setting(settings, "PAGE_MODEL")


def get_hero_model():
    return get_model_from_setting(settings, "HERO_MODULE_MODEL")


def get_header_model():
    return get_model_from_setting(settings, "HEADER_MODULE_MODEL")


def get_two_column_model():
    return get_model_from_setting(settings, "TWO_COLUMN_MODEL")


def get_icon_list_model():
    return get_model_from_setting(settings, "ICON_LIST_MODEL")


def get_icon_item_model():
    return get_model_from_setting(settings, "ICON_LIST_ITEM_MODEL")


def get_faq_model():
    return get_model_from_setting(settings, "FAQ_MODULE_MODEL")


def get_faq_item_model():
    return get_model_from_setting(settings, "FAQ_ITEM_MODEL")


def get_location_model():
    return get_model_from_setting(settings, "LOCATION_MODULE_MODEL")


def get_location_item_model():
    return get_model_from_setting(settings, "LOCATION_ITEM_MODEL")


def get_image_gallery_model():
    return get_model_from_setting(settings, "GALLERY_MODULE_MODEL")


def get_gallery_item_model():
    return get_model_from_setting(settings, "GALLERY_ITEM_MODEL")


def get_carousel_model():
    return get_model_from_setting(settings, "CAROUSEL_MODULE_MODEL")


def get_carousel_item_model():
    return get_model_from_setting(settings, "CAROUSEL_ITEM_MODEL")



import pathlib
import yaml

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction
from django.db.models import Model

from scarlet.pagebuilder.base_models import BasePage
from scarlet.pagebuilder import get_page_model


class Command(BaseCommand):
    help = "Generate blank pages with modules from YML schema"

    def add_arguments(self, parser):
        parser.add_argument(
            "--file",
            help="Path to file with YAML schema",
            type=pathlib.Path,
        )
        parser.add_argument(
            "--allow-duplicates",
            action="store_true",
            help="Allows to create pages with existing slug",
        )

    @transaction.atomic
    def handle(self, *args, **options):
        Page: BasePage = get_page_model()

        filename = (
            options.get("file") or
            settings.APPS_DIR / "pages/scaffold-schema.yml"
        )
        with open(filename, "r") as f:
            schema = yaml.load(f, Loader=yaml.Loader)

        # Simple validation
        for page_name, data in schema.items():
            keys = data.keys()
            if 'slug' not in keys:
                return self.stdout.write(self.style.ERROR(
                    f"No 'slug' in {page_name}"
                ))

            if not options.get("allow_duplicates"):
                if Page.objects.filter(slug=data["slug"]).exists():
                    return self.stdout.write(self.style.ERROR(
                        f"Page with slug `{data['slug']}` already exists"
                    ))

            for module in data.get("modules", []):
                try:
                    apps.get_model("pages", f"{module}Module")
                except LookupError:
                    return self.stdout.write(self.style.ERROR(
                        f"Invalid module: {module}"
                    ))

        for page_name, data in schema.items():
            modules = data.pop("modules")
            page = get_page_model().objects.create(title=page_name, **data)
            for module in modules:
                model: Model = apps.get_model("pages", f"{module}Module")
                model.objects.create(page=page)

        self.stdout.write(self.style.SUCCESS("Success"))



import warnings

from scarlet.utils.assets import get_image_urls as new_get_image_urls


def get_image_urls(obj, field, sizes=None):
    warnings.warn("get_image_urls has been moved to scarlet.utils.assets "
                  "and will be removed from this module in a future version.",
                  category=DeprecationWarning, stacklevel=2)
    return new_get_image_urls(obj, field, sizes)



"""
This file can be copied into a project's app folder as models.py for
default implementation of disclosure models.
"""
from django.db import models
from scarlet.disclosures.models import BaseGlossaryTerm, BaseGlossaryDisclosure
from scarlet.versioning.models import Cloneable


class GlossaryManager(models.Manager):

    def get_queryset(self):
        return super().get_queryset().prefetch_related("glossarydisclosures")


class GlossaryTerm(Cloneable, BaseGlossaryTerm):
    objects = GlossaryManager

    def __str__(self):
        return self.keyword


class GlossaryDisclosure(Cloneable, BaseGlossaryDisclosure):

    def __str__(self):
        return self.title



import re

from django.db import models
from scarlet.cms.fields import HTMLTextField

from . import settings, get_glossary_term_model


class BaseGlossaryTerm(models.Model):
    _clone_related = ("glossarydisclosures",)

    keyword = models.CharField(
        max_length=255,
        unique=True,
        help_text="The word or short phrase that will appear with a disclosure. Case insensitive.",
    )

    class Meta:
        abstract = True


class BaseGlossaryDisclosure(models.Model):
    term = models.ForeignKey(settings.GLOSSARY_TERM_MODEL,
                             on_delete=models.PROTECT,
                             related_name="glossarydisclosures")
    title = models.CharField(max_length=255)
    body = HTMLTextField(config="useMinimal")

    class Meta:
        abstract = True


class DisclosureMixin:
    """
    Using this mixin allows you to add disclosure markup to any string field on
    a model. The mixin assumes there is a model imported called GlossaryTerm but
    you can change that by overriding the get_glossary_model function. A glossary
    model must have at least an id field, and a keyword field.

    On model:
        Use this mixin, set disclosure_fields to be a list of all field names on the model that should
        get annotated, and add self.add_disclosures() call to save() method.

    For serialization:
        In order to pull the disclosures that go with the marked up text fields call
        get_disclosures_from_fields, this will return a list of Glossary model objects
        for you to serialize.
    """
    disclosure_fields = []

    @staticmethod
    def _span_wrap(matchobj):
        return "<span data-disclosure-id={{term_id}}>{keyword}</span>".format(
            keyword=matchobj.group(0)
        )

    def _get_disclosed_text(self, term, text):
        escaped = re.escape(term.keyword)
        search = re.compile(rf'(?<!<span data-disclosure-id={term.id}>){escaped}',
                            re.IGNORECASE)
        wrapped_str = search.sub(self._span_wrap, text)
        return wrapped_str.format(term_id=term.id)

    def add_disclosures(self):
        for field in self.disclosure_fields:
            value = getattr(self, field, None)
            if value:
                glossary = self.get_glossary()
                for term in glossary:
                    if term.keyword.lower() in value.lower():
                        value = self._get_disclosed_text(term, value)
                setattr(self, field, value)

    def _get_ids_from_text(self, text):
        return set(re.findall(r'(?<=span data-disclosure-id=)\d', text))

    def get_disclosures_from_fields(self):
        term_ids = []
        for field in self.disclosure_fields:
            value = getattr(self, field, None)
            if value:
                term_ids += self._get_ids_from_text(value)
        return self.get_glossary_model().objects.filter(id__in=term_ids)

    def get_glossary_model(self):
        return get_glossary_term_model()

    def get_glossary(self):
        return self.get_glossary_model().objects.all()



from django.conf import settings

GLOSSARY_DISCLOSURE_MODEL = getattr(settings, "GLOSSARY_DISCLOSURE_MODEL", "common.GlossaryDisclosure")
GLOSSARY_TERM_MODEL = getattr(settings, "GLOSSARY_TERM_MODEL", "common.GlossaryTerm")



from types import ModuleType

from django.apps import apps as djapps
from django.core.exceptions import ImproperlyConfigured


def get_model_from_setting(settings: ModuleType, setting_name: str):
    """
    Get the model from a settings in the provided settings module.
    :param settings: A settings module containing the setting.
    :param setting_name: The name of the settings (e.g. "PAGE_MODEL")
    :return: The model class
    """

    try:
        app_label, model_name = getattr(settings, setting_name).rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            f"{setting_name} must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)



from django.forms import inlineformset_factory

from scarlet.cms import forms, bundles, views
from scarlet.disclosures import get_glossary_term_model, get_glossary_disclosure_model


class DisclosureInlineFormset(forms.LazyFormSetFactory):
    def __init__(self):
        super().__init__(
            inlineformset_factory,
            get_glossary_term_model(),
            get_glossary_disclosure_model(),
            can_order=False,
            can_delete=True,
        )


class GlossaryBundle(bundles.Bundle):
    dashboard = (("main",),)

    edit = views.FormView(formsets={"Disclosures": DisclosureInlineFormset()})
    add = edit

    class Meta:
        model = get_glossary_term_model()



from scarlet.utils.lookups import get_model_from_setting
from . import settings


def get_glossary_term_model():
    return get_model_from_setting(settings, "GLOSSARY_TERM_MODEL")


def get_glossary_disclosure_model():
    return get_model_from_setting(settings, "GLOSSARY_DISCLOSURE_MODEL")



from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from scarlet.disclosures import get_glossary_term_model
from scarlet.disclosures.openapi import DisclosureSchema


class GlossarySerializer(serializers.ModelSerializer):
    disclosure = serializers.SerializerMethodField()

    @extend_schema_field(DisclosureSchema)
    def get_disclosure(self, queryset):
        resp = {}
        for obj in queryset:
            resp[obj.id] = []
            for disclosure in obj.glossarydisclosures.all():
                resp[obj.id].append(
                    {"title": disclosure.title, "body": disclosure.body}
                )
        return resp

    class Meta:
        model = get_glossary_term_model()
        fields = ("id", "disclosure")



from django.db.models import Manager
from django.conf import settings as dj_settings
from rest_framework import serializers

from scarlet.utils.assets import ImageSerializer
from . import (
    settings,
    get_page_model,
    get_hero_model,
    get_header_model,
    get_two_column_model,
    get_icon_list_model,
    get_icon_item_model,
    get_faq_model,
    get_faq_item_model,
    get_location_model,
    get_location_item_model,
    get_image_gallery_model,
    get_gallery_item_model,
    get_carousel_model,
    get_carousel_item_model,
)

Page = get_page_model()
HeroModule = get_hero_model()
HeaderModule = get_header_model()
TwoColumnModule = get_two_column_model()
IconListModule = get_icon_list_model()
IconListItem = get_icon_item_model()
FAQModule = get_faq_model()
FAQItem = get_faq_item_model()
LocationModule = get_location_model()
LocationItem = get_location_item_model()
ImageGalleryModule = get_image_gallery_model()
GalleryImage = get_gallery_item_model()
CarouselModule = get_carousel_model()
CarouselItem = get_carousel_item_model()


class HeroModuleSerializer(serializers.ModelSerializer):
    image = ImageSerializer(sizes=settings.WIDE_IMAGE_SIZES)

    class Meta:
        model = HeroModule
        fields = (
            "kind",
            "image",
            "header",
            "subheader",
            "button_url",
            "button_text",
            "style",
            "behavior",
        )


class TwoColumnModuleSerializer(serializers.ModelSerializer):
    left_image = ImageSerializer(sizes=settings.SQUARE_IMAGE_SIZES)
    right_image = ImageSerializer(sizes=settings.SQUARE_IMAGE_SIZES)

    class Meta:
        model = TwoColumnModule
        fields = (
            "kind",
            "left_image",
            "left_header",
            "left_description",
            "left_button_url",
            "left_button_text",
            "right_image",
            "right_header",
            "right_description",
            "right_button_url",
            "right_button_text",
            "style",
            "behavior",
        )


class HeaderModuleSerializer(serializers.ModelSerializer):
    class Meta:
        model = HeaderModule
        fields = (
            "kind",
            "header",
            "subheader",
            "style",
            "behavior",
        )


class IconListItemSerializer(serializers.ModelSerializer):
    icon = ImageSerializer()

    class Meta:
        model = IconListItem
        fields = (
            "kind",
            "icon",
            "text",
        )


class IconListModuleSerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = IconListModule
        fields = (
            "kind",
            "title",
            "subtitle",
            "footnote",
            "items",
            "style",
            "behavior",
        )


class FAQItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = FAQItem
        fields = (
            "kind",
            "question",
            "answer",
        )


class FAQModuleSerializer(serializers.ModelSerializer):
    items = FAQItemSerializer(many=True)

    class Meta:
        model = FAQModule
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class LocationItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = LocationItem
        fields = (
            "kind",
            "title",
            "address1",
            "address2",
            "city",
            "region",
            "zip_code",
            "country",
            "hours",
            "phone_number",
        )


class LocationModuleSerializer(serializers.ModelSerializer):
    items = LocationItemSerializer(many=True)

    class Meta:
        model = LocationModule
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class GalleryImageSerializer(serializers.ModelSerializer):
    image = ImageSerializer(sizes=settings.GALLERY_IMAGE_SIZES)

    class Meta:
        model = GalleryImage
        fields = (
            "kind",
            "image",
        )


class ImageGallerySerializer(serializers.ModelSerializer):
    items = GalleryImageSerializer(many=True)

    class Meta:
        model = ImageGalleryModule
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class CarouselItemSerializer(serializers.ModelSerializer):
    image = ImageSerializer(sizes=settings.CAROUSEL_IMAGE_SIZES)

    class Meta:
        model = CarouselItem
        fields = (
            "kind",
            "title",
            "subtitle",
            "image",
            "button_url",
            "button_text",
        )


class CarouselModuleSerializer(serializers.ModelSerializer):
    items = CarouselItemSerializer(many=True)

    class Meta:
        model = CarouselModule
        fields = (
            "kind",
            "button_url",
            "button_text",
            "items",
            "style",
            "behavior",
        )


class OpenGraphWebsiteSerializer(serializers.Serializer):
    kind = serializers.ReadOnlyField(default="website")
    url = serializers.CharField(default=settings.SITE_URL)
    locale = serializers.CharField(default="en-US")
    title = serializers.CharField(default="Wonderful")
    description = serializers.CharField(source="seo_description", default="")
    image = ImageSerializer(source="og_image", default={})
    site_name = serializers.CharField(default="Wonderful")
    card = serializers.ReadOnlyField(default="summary_large_image")


class PageHeadModuleSerializer(serializers.Serializer):
    def __init__(self, instance=None, **kwargs):
        instance = {**instance, 'open_graph': instance}
        super().__init__(instance, **kwargs)

    kind = serializers.ReadOnlyField(default="head")
    locale = serializers.CharField(default="en-US")
    title = serializers.CharField(default="Wonderful")
    description = serializers.CharField(source="seo_description", default="")
    keywords = serializers.CharField(default="")
    url = serializers.CharField()
    gtm = serializers.CharField(default="gtm-Test")
    open_graph = OpenGraphWebsiteSerializer()


class PageListSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        iterable = data.all() if isinstance(data, Manager) else data
        return [self.child.to_representation(item, many=True) for item in iterable]


class PageSerializer(serializers.ModelSerializer):
    """
    Serializer to serialize Page model or any module using flexible module structure.
    Creates response with `head` object as first item in list, followed by all modules
    serialized based on the serializer_map.
    """

    og_image = ImageSerializer()

    def _get_modules(self, obj) -> list:
        modules = []
        for module_type in self.Meta.module_fields:
            modules += getattr(obj, module_type).all()
            # TODO: Test if modules.insert() for each in all() is faster than current
        return sorted(modules, key=lambda x: x.order,)

    def _module_to_representation(self, module) -> str:
        return self.Meta.serializer_map.get(type(module))(module).data

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs["child"] = cls()
        return PageListSerializer(*args, **kwargs)

    def to_representation(self, instance, many=False):
        orig_rep = super().to_representation(instance)

        request = self.context.get("request", None)
        url = request.build_absolute_uri() if request else dj_settings.SITE_URL

        head_module = PageHeadModuleSerializer({**orig_rep, 'url': url}).data
        new_rep = {"content": [head_module]}

        if many:
            body = {}
            for key in self.Meta.body_fields:
                body[key] = orig_rep.get(key)

            new_rep["content"].append(body)
        else:
            modules = self._get_modules(instance)
            for mod in modules:
                new_rep["content"].append(self._module_to_representation(mod))

        return new_rep

    class Meta:
        model = Page
        body_fields = (
            "title",
            "slug",
            "seo_description",
            "keywords",
            "og_image",
            "site_name",
        )
        fields = body_fields
        module_fields = (
            "page_hero_modules",
            "page_two_column_modules",
            "page_icon_list_modules",
            "page_header_modules",
            "page_faq_modules",
            "page_location_modules",
            "page_gallery_modules",
            "page_carousel_modules",
        )
        serializer_map = {
            HeroModule: HeroModuleSerializer,
            TwoColumnModule: TwoColumnModuleSerializer,
            IconListModule: IconListModuleSerializer,
            HeaderModule: HeaderModuleSerializer,
            FAQModule: FAQModuleSerializer,
            LocationModule: LocationModuleSerializer,
            ImageGalleryModule: ImageGallerySerializer,
            CarouselModule: CarouselModuleSerializer,
        }


class ObjectPageSerializer(serializers.ModelSerializer):
    """
    Serializes an object similar to the PageSerializer but without module fields
    or serializer map needed for variable modules. Fields should be registered
    like a standard DRF serializer and added to the meta body_fields.
    """

    og_image = ImageSerializer(required=False)

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs["child"] = cls()
        return PageListSerializer(*args, **kwargs)

    def to_representation(self, instance, many=False):
        orig_rep = super().to_representation(instance)

        request = self.context.get("request", None)
        url = request.build_absolute_uri() if request else dj_settings.SITE_URL

        head_module = PageHeadModuleSerializer({**orig_rep, 'url': url}).data
        new_rep = {"content": [head_module]}

        if many:
            body = {}
            for key in self.Meta.body_fields:
                body[key] = orig_rep.get(key)

            new_rep["content"].append(body)
        else:
            for key in self.Meta.body_fields:
                if not orig_rep.get(key):
                    continue

                new_rep["content"].append(
                    {"kind": key, "items": orig_rep.get(key),}
                )

        return new_rep

    class Meta:
        model = Page
        body_fields = (
            "title",
            "slug",
            "description",
            "keywords",
            "og_image",
        )
        fields = body_fields



try:
    from ..cms import views
except ValueError:
    from cms import views


class ItemsListView(views.ListView):
    default_template = "cms/items_list.html"

    def get_queryset(self, **filter_kwargs):
        module_id = self.request.GET.get("module_id")
        if module_id:
            filter_kwargs.update({"module_id": module_id})
        qs = super().get_queryset(**filter_kwargs)
        return qs.distinct()



from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers


def get_image_urls(obj, field, sizes=None):
    if sizes is None:
        sizes = {'base': None}

    images = {}
    for size in sizes:
        img_field = f"{field}_urls"
        try:
            images[size] = (
                getattr(obj, img_field).get(size) if getattr(obj, img_field) else None
            )
        except:
            images[size] = ""
    return images


class BaseImageSerializer(serializers.Serializer):
    altText = serializers.CharField()

    def __init__(self, *args, **kwargs):
        """TODO: Figure out a way to get/generate schema from sizes param"""
        self.sizes = kwargs.pop('sizes', None)
        super().__init__(*args, **kwargs)

    def to_representation(self, instance):
        module = self.parent.instance
        images_json = get_image_urls(module, self.field_name, self.sizes)
        images_json["altText"] = getattr(module, f"{self.field_name}_alt_text", None)
        return images_json


class ImageSerializer(BaseImageSerializer):
    base = serializers.CharField(required=False)



from rest_framework import serializers


class DisclosureItemSchema(serializers.Serializer):
    title = serializers.CharField()
    body = serializers.CharField()


class DisclosureSchema(serializers.Serializer):
    disclosure = DisclosureItemSchema()
